from typing import List, LiteralString, Dict

class Constructor:
    """
    A class for constructing new filenames based on extracted data.

    Attributes:
        data_in : List
            A list of one or more 'extractions' from the Extractor() class.
        success : List
            A list of successful constructions.
        failure : List
            A list of unsuccessful constructions.
        auto : bool
            A flag for whether or not to perform auto-construction.
        constructor : LiteralString
            A string representing the constructor style to use.
    """
    def __init__(self, data_list: List, constructor: LiteralString, auto: bool = False) -> None:
        """
        Load the constructor with the data, a constructor choice and optional auto option.

        Parameters:
            data_list (List): A list of datasets. Each dataset should be a dictionary of extracted data 
                generated by the Extractor() class.
            constructor (LiteralString): A string that sets the constructor style to use. Options are 'custom', 
                'simple', or 'block'.
            auto (bool): An optional boolean that sets whether or not to use auto-construction. Default is False.

        Returns:
            None
        """
        self.data_in = data_list
        self.success = []
        self.failure = []
        self.auto = auto
        self.constructor = constructor

    # return successful constructions
    def get_success(self):
        """
        Returns the list of successful constructions.

        Returns
        -------
            List: A list of successful constructions.
        """
        if self.success:
            return self.success

    # return unsuccessful constructions
    def get_failure(self):
        """
        Returns the list of unsuccessful constructions.

        Returns:
            List: A list of unsuccessful constructions.
        """
        if self.failure:
            return self.failure

    # construction from a list of one or more dicts or json datasets
    # constructions occur for extractions with a status of True indicating
    # that the file info was obtained successfully from Extractor()
    # constructions can be obtained through the get_success method,
    # files that wont be constructed can be obtained through the get_failure method
    def from_dict(self):
        """
        Construction from a list of one or more dicts or json datasets. Constructions occur for extractions with a 
        status of True indicating that the file info was obtained successfully from Extractor(). Constructions can be 
        obtained through the get_success method, files that wont be constructed can be obtained through the 
        get_failure method.

        Parameters:
            None

        Returns:
            None
        """
        # cycle through extractions within the data list
        for _data in self.data_in:

            # if a dataset within the list is not a dict, then raise TypeError
            if not isinstance(_data, Dict):
                raise TypeError("All data in list should be dicts")

            # get the `status` of the current extraction
            status = _data["status"]
            # get the original filename of the current extraction
            original_file = _data["original_file"]

            # if the extraction status = True aka successful, then get the values
            # from the extraction data and send them through the run method - the 
            # resulting tuple will be appended to the `success` property
            if status:
                artist = _data["artist"]
                title = _data["title"]
                features = _data["features"]
                misc = _data["misc"]
                filetype = _data["filetype"]
                youtube_id = _data["youtube_id"]

                self.success.append(
                    self.run(
                        original_file,
                        artist,
                        title,
                        features,
                        misc,
                        filetype,
                        youtube_id
                    )
                )
            if not status:
                self.failure.append(original_file)


    def run(
        self,
        original_file: LiteralString,
        artist: LiteralString,
        title: LiteralString,
        features: LiteralString,
        misc: LiteralString,
        filetype: LiteralString,
        youtube_id: LiteralString
    ) -> tuple:
        """
        Singular construction from extracted values, typically will be called by the other methods within Constructor, 
        returns a tuple containing the original filename and a list containing its new formatting options, if auto is 
        true then the most detailed option is returned with the filename instead of a list of options.

        Parameters:
            original_file (LiteralString): The original filename to be formatted.
            artist (LiteralString): The artist name extracted from the file.
            title (LiteralString): The song title extracted from the file.
            features (LiteralString): Any features of the song, if any.
            misc (LiteralString): Any additional information to be included in the filename, if any.
            filetype (LiteralString): The filetype of the file.
            youtube_id (LiteralString): The youtube id, if available.

        Returns:
            tuple: A tuple containing the original filename and a list of options for the formatted filename. If 
                auto is True, the most detailed option is returned with the filename instead of a list of options.
                Tuple[str, List] or Tuple[str, str]
        """

        if self.constructor == "custom":
            # have user select desired values
            # have user select layout from options generated from selected values
            pass

        if self.constructor == "simple":
            options = [
                f"{artist} - {title}.{filetype}",
                f"{artist} - {title} [{youtube_id}].{filetype}",
                f"{artist} - {title} ({misc}).{filetype}",
                f"{artist} - {title} ft. {features}.{filetype}",
                f"{artist} - {title} ft. {features} ({misc}).{filetype}",
                f"{artist} - {title} ft. {features} [{youtube_id}].{filetype}",
                f"{artist} - {title} ft. {features} ({misc}) [{youtube_id}].{filetype}",
            ]

        elif self.constructor == "enclosed":
            options = [
                f"({artist})({title}).{filetype}",
                f"({artist})({title})({misc}).{filetype}",
                f"({artist})({title})({features}).{filetype}",
                f"({artist})({title})({youtube_id}).{filetype}",
                f"({artist})({title})({features})({misc}).{filetype}",
                f"({artist})({title})({features})({youtube_id}).{filetype}",
                f"({artist})({title})({features})({misc})({youtube_id}).{filetype}"
            ]

        if self.auto:
            return (original_file, options[-1])

        return (original_file, options)



from audiodotturn import VERSION

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "AudioDotTurn"
copyright = "2023, tairenfd"
author = "tairenfd"
release = VERSION

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.githubpages",
    "sphinx.ext.autosummary",
    'sphinx_search.extension',
    'myst_parser',
]

source_suffix = ['.rst', '.md']
templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



from typing import List, Dict, Any
from io import StringIO
from datetime import datetime
from rich.console import Console
from audiodotturn.config import ConfigUser
from audiodotturn.construct import Constructor
from audiodotturn.extract import Extractor
from audiodotturn.database import Database


class AudioDotTurn:
    def __init__(self, config_path: str = None, db_path: str = None):
        self.config = ConfigUser(config_path)
        self.extractor = Extractor(self.config.exts, self.config.output_opts)
        self.database = Database(db_path or self.config.db_path)
        self.current_data = None
        self.constructor = None

    def extract_files(self, files: List[str], output_format: str = "dict") -> List[Any]:
        """
        Extracts metadata from multiple audio files and returns a list of dictionaries which
        contain the data or a list of the data in the chosen format.
        """
        self.current_data = self.extractor.extract_complex_list(files, output_format)
        return self.current_data

    def extract_file(self, file: str, opt: str = "dict") -> List[Any]:
        """
        Extracts metadata from a single audio file and returns a list containing a single
        dictionary containing the data.
        """
        self.extractor.complex_extract(file)
        self.current_data = self.extractor.get_extraction(opt)
        return [self.current_data]

    def update_database(self, data: List[Dict] = None) -> None:
        """
        Updates the database with data extracted from audio files.
        """
        self.database.create_database()
        self.database.create_tables()
        return self.database.update_database(data or self.current_data)
    
    def produce_construct_report(self, results, console, args):
        """
        Produce report of constructions. Options are html, svg, txt, or console.

        ONLY FOR USE WITH CLI CLIENT        
        """
        success = 0
        failure = 0

        report_type = input(
            "\nHow would you like to process the constructions(s)?\nOutput will be sent to working directory named 'construct_report.[ext]' if a file.\nOptions: html, text, svg, console\nFor none press enter. "
        )
        match report_type.lower().strip():
            case "html":
                report_type = "html"
            case "svg":
                report_type = "svg"
            case "text":
                report_type = "text"
            case "console":
                report_type = "console"
            case _:
                report_type = None

        if report_type is not None:
            if report_type == "console":
                if not args.auto:
                    if results["success"]:
                        if args.multi:
                            for result in results["success"]:
                                console.print(f"\n[cyan]Original File: [magenta]{result[0]}\n")
                                console.print("Options created:", style="cyan")
                                for option in result[1]:
                                    console.print(option, style="success")
                                success += 1
                        else:
                            console.print(f"\nOriginal File: {results['success'][0][0]}\n")
                            console.print("Options created:", style="cyan")
                            for option in results["success"][0][1]:
                                console.print(option, style="success")
                            success += 1
                        console.print('\n')
                else:
                    if results["success"]:
                        if args.multi:
                            for result in results["success"]:
                                console.print(f'[magenta]"{result[0]}" => [success]"{result[1]}"\n')
                            success += 1
                        else:
                            console.print(f'[magenta]{results["success"][0][0]} => [success]{results["success"][0][1]}', style="success")
                            success += 1

                        console.print('\n')

                if results["failure"]:
                    for result in results["failure"]:
                        failure += 1
                        console.print(f"\nFailed to construct: {result}\n", style="failure")
            else:
                record_console = Console(record=True, stderr=True, file=StringIO())
                if not args.auto:
                    if results["success"]:
                        if args.multi:
                            for result in results["success"]:
                                record_console.print(f"\n[cyan]Original File: [magenta]{result[0]}\n")
                                record_console.print("Options created:", style="cyan")
                                for option in result[1]:
                                    record_console.print(option, style="green")
                                success += 1
                        else:
                            record_console.print(f"\nOriginal File: {results['success'][0][0]}\n")
                            record_console.print("Options created:", style="cyan")
                            for option in results["success"][0][1]:
                                record_console.print(option, style="green")
                            success += 1
                        record_console.print('\n')
                else:
                    if results["success"]:
                        if args.multi:
                            for result in results["success"]:
                                record_console.print(f'[magenta]"{result[0]}" => [green]"{result[1]}"\n')
                                success += 1
                        else:
                            record_console.print(f'[magenta]{results["success"][0][0]} => [success]{results["success"][0][1]}', style="success")
                            success += 1

                        record_console.print('\n')

                if results["failure"]:
                    for result in results["failure"]:
                        failure += 1
                        record_console.print(f"\nFailed to construct: {result}\n", style="red")

                match report_type:
                    case "html":
                        record = record_console.export_html()
                    case "svg":
                        record = record_console.export_svg()
                    case "txt":
                        record = record_console.export_text()
                    case _:
                        record = False

                if record:
                    with open(f"construct_report.{report_type}", "wt") as report:
                        report.write(record)

        return success, failure


    def produce_extract_report(self, extractions: List[Dict], console):
        """
        Produce report of extractions list. Options are html, svg, txt, or console.

        ONLY FOR USE WITH CLI CLIENT
        """

        report_type = input(
            "\nHow would you like to process the extraction(s)?\nOutput will be sent to working directory named 'extract_report.[ext]' if a file.\nOptions: html, text, svg, console\nFor none press enter. "
        )
        match report_type.lower().strip():
            case "html":
                report_type = "html"
            case "svg":
                report_type = "svg"
            case "text":
                report_type = "text"
            case "console":
                report_type = "console"
            case _:
                report_type = None


        if report_type is not None:
            if report_type == "console":
                console.print('Extracted Info:', style="cyan")


                for extracted in extractions:
                    for key, value in extracted.items():
                        console.print(key, ':', value, style="success")
                    
                    console.print('\n')
            else:
                record_console = Console(record=True, stderr=True, file=StringIO())
                record_console.rule('Extracted Info:', style="cyan")

                for extracted in extractions:
                    for key, value in extracted.items():
                        record_console.print(key, ':', value, style="bold green")
                    
                    record_console.print('\n')
                record_console.rule(f"Report Generated {datetime.now().ctime()}")


                match report_type:
                    case "html":
                        record = record_console.export_html()
                    case "svg":
                        record = record_console.export_svg()
                    case "txt":
                        record = record_console.export_text()
                    case _:
                        record = False

                if record:
                    with open(f"extract_report.{report_type}", "wt") as report:
                        report.write(record)
        
        else:
            console.print("No report generated.", style="yellow")

    def get_all_artists(self) -> List[Dict]:
        """
        Returns a list of all artists in the database.
        """
        return self.database.get_all_artists()

    def get_artist_by_id(self, artist_id: int) -> List[Dict]:
        """
        Returns a list of all artists in the database.
        """
        return self.database.get_artist_by_id(artist_id)

    def get_all_artists_and_songs(self) -> Dict:
        """
        Returns a dict of all artists and their songs in the database.
        """
        songs = {}
        artists = self.database.get_all_artists()
        for artist in artists:
            songs[artist["name"]] = self.database.get_songs_by_artist_by_id(artist["artist_id"])
        return songs

    def get_songs_by_artist(self, artist_id: int) -> List[Dict]:
        """
        Returns a list of all songs by a given artist.
        """
        return self.database.get_songs_by_artist_by_id(artist_id)

    def get_song_by_id(self, song_id: int) -> Dict:
        """
        Returns a song with the given ID.
        """
        return self.database.get_song_by_id(song_id)

    def construct(self, constructor: str, data: List[Dict] = None, auto: bool = False) -> Dict:
        """
        Constructs new audio files from the data in the database and returns a dictionary containing the
        filenames and the options used to construct them.
        """
        data = data or self.current_data
        if constructor not in self.config.constructors:
            raise TypeError(f"constructor {constructor} does not exist")
        self.constructor = Constructor(data, constructor, auto)
        self.constructor.from_dict()
        return {
            "success": self.constructor.get_success(),
            "failure": self.constructor.get_failure()
        }



import re
import json
import os
from typing import Union, List, LiteralString, Any, Dict
import yaml

class Extractor:
    """
    A class for extracting metadata from file names.

    Attributes:
        exts: Tuple of strings
            The file extensions that the extractor class will attempt to extract metadata from.
        
        output_opts: List of strings
            The available output format options for the extracted data.
        
        extracted_data: Dict[str, Union[str, Any]]
            The extracted metadata from the last extraction.
    """
    # initialize extractor instance with optional extension list
    def __init__(self, exts: List, output_opts = List):
        """
        Initializes an instance of the Extractor class with an optional list of file extensions.

        Parameters:
            exts: list of exts to format
            output_opts: available output opts, this should be the opts in the program defaults.
        """

        self.exts = tuple(exts)
        self.output_opts = output_opts
        self.extracted_data = None

    def get_extraction(self, opt: str = "dict"):
        """
        Returns the last extraction performed in the extraction instance
        in a specified format.

        Parameters:
            opt: str, optional
                The desired format for the extracted data. The default format is "dict". The available format options are:
                "dict", "json", "yaml", "str", "list", "keys", and "values".

        Returns:
            The extracted data in the specified format. The format will depend on the specified `opt` parameter.

        Raises:
            None
        """
        if opt and opt.lower() in self.output_opts and self.extracted_data:
            match opt:
                case "json":
                    return json.dumps(self.extracted_data)
                case "yaml":
                    return yaml.dump(self.extracted_data)
                case "dict":
                    return self.extracted_data
                case "str":
                    return ' '.join([str(value) for value in self.extracted_data.values()])
                case "list":
                    return list(self.extracted_data)
                case "keys":
                    return list(self.extracted_data.keys())
                case "values":
                    return list(self.extracted_data.values())
                case _:
                    return self.extracted_data
        else:
            raise UserWarning("Extracted data is empty")

    # unsuccessful extraction
    # sets current instances most current extraction via
    # extracted-data property
    def false_extract(self, _file: LiteralString) -> Dict[str, Union[str, Any]]:
        """
        Sets the current extraction as an unsuccessful extraction, unsucceful extractions
        simply contain the original file name and a False status, all metadata values are set
        to None

        Parameters:
            _file: str
                The file name to extract metadata from.

        Returns:
            A dictionary containing the unsuccessful extraction data.

        Raises:
            TypeError:
                If `_file` is not a string.
        """
        if not isinstance(_file, str):
            raise TypeError("file must be a str literal")

        self.extracted_data = {
            "original_file": _file,
            'artist': None,
            'title': None,
            'features': None,
            'misc': None,
            'youtube_id': None,
            'filetype': None,
            "status": False
        }

    # successful extraction
    # sets current instances most current extraction via
    # extracted-data property
    def true_extract(self, info: List[Any]) -> Dict[str, Union[str, Any]]:
        """
        Sets the most current extraction as successful and sets the `extracted_data` attribute with the extracted data.

        Parameters:
            info: List
                A list containing the extracted metadata.

        Returns:
            A dictionary containing the successful extraction data.

        Raises:
            TypeError:
                If `info` is not a list.
        """
        if not isinstance(info, List):
            raise TypeError("info must be a list of 7 values")

        self.extracted_data = {
            "original_file": info[0],
            'artist': info[1],
            'title': info[2],
            'features': info[3],
            'misc': info[4],
            'youtube_id': info[5],
            'filetype': info[6],
            "status": info[7]            
        }


    # simple extraction for if you already know the formatting to an extent
    def simple_extract(self, _file: LiteralString) -> List[Any]:
        """
        Extracts information from a filename in a specific format and returns the extracted
        data as a list containing eight values: the original filename, the artist info, title
        info, features info, misc info, youtube_id info, the file extension, and the extractions
        status value (True or False depending on whether data was sent to the `true_extract` method
        or the `false_extract` method by the Extractor).

        If the filename info is extracted successfully, the method uses the `true_extract` method
        to set the current extract data. If unsuccessful, the method uses the `false_extract` method
        to set the current extract data. Current extract data always contains the last extract
        within the current `Extractor()` instance. Extracts are tuples containing eight values:
        the original filename, the artist info, title info, features info, misc info, youtube_id
        info, the file extension, and the extractions status value which will be True or False
        depending on whether data was sent to the `true_extract` method or the `false_extract`
        method.

        Parameters:
            _file (LiteralString): The filename to extract information from.

        Returns:
            List[Any]: The extracted information in the form of a list containing eight values:
                the original filename, the artist info, title info, features info, misc info,
                youtube_id info, the file extension, and the extractions status value.
        """

        if not isinstance(_file, str):
            raise TypeError("file must be a str literal")

        _file = os.path.basename(_file)

        if not _file.endswith(self.exts):
            return self.false_extract(_file)

        format_check = re.search(
            r"\[(.+?)\][ ]?\[(.+?)\][ ]?\[(.+?)\][ ]?\[(.+?)\][ ]?\[(.+?)\][ ]?\.(\w+)$", _file
        ) or re.search(
            r"^(.+?)[ ]?-[ ](.+?) ft\. (.+?) \((.+?)\) \[(.+?)\]\.(\w+)$|^(.+?)[ ]-[ ]?(.+?)ft\.(.+?)\((.+?)\) \[(.+?)\]\.(\w+)$",
            _file
        ) or re.search(
            r"\((.+?)\)[ ]?\((.+?)\)[ ]?\((.+?)\)[ ]?\((.+?)\)[ ]?\((.+?)\)[ ]?\.(\w+)$", _file
        )

        if format_check:

            return self.true_extract([
                    _file,
                    format_check[1],
                    format_check[2],
                    format_check[3],
                    format_check[4],
                    format_check[5],
                    format_check[6],
                    True
                ])

        format_check = re.search(
            r"\[(.+?)\][ ]?\[(.+?)\][ ]?\[(.+?)\][ ]?\[(.+?)\]\.(\w+)$",
            _file
        ) or re.search(
            r"^(.+?)[ ]?-[ ](.+?)ft\.(.+?)\((.+?)\)\.(\w+)$|^(.+?)[ ]-[ ]?(.+?)ft\.(.+?)\((.+?)\)\.(\w+)$",
            _file
        ) or re.search(
            r"\((.+?)\)[ ]?\((.+?)\)[ ]?\((.+?)\)[ ]?\((.+?)\)\.(\w+)$",
            _file
        )

        if format_check:

            return self.true_extract([
                    _file,
                    format_check[1],
                    format_check[2],
                    format_check[3],
                    format_check[4],
                    None,
                    format_check[5],
                    True
            ])

        format_check = re.search(
            r"\[(.+?)\][ ]?\[(.+?)\][ ]?\[(.+?)\]\.(\w+)$", _file
        ) or re.search(
            r"\((.+?)\)[ ]?\((.+?)\)[ ]?\((.+?)\)\.(\w+)$", _file
        )

        if format_check:
            return self.true_extract([
                    _file,
                    format_check[1],
                    format_check[2],
                    format_check[3],
                    None,
                    None,
                    format_check[4],
                    True
            ])
       
        if not format_check:
            return self.false_extract(_file)

    # 'complex' extraction for any filename
    # if filename info is extracted successfully, then use true extract method to set current
    # extract data
    # if unsuccessful, use the false extract method to set current extract data
    # current extract data will always contain the last extract within the current
    # Extractor() instance
    # extracts are tuples containing 8 values, the original filename, the artist info, title
    # info, features info, misc info, youtube_id info, the file extension, and the extractions
    # status value which will be True or False depending on if data was sent to true extract method
    # or false extract method
    def complex_extract(self, _file: LiteralString) -> List[Any]:
        """
        Extracts information from a filename and returns the extracted data as a list containing
        eight values: the original filename, the artist info, title info, features info, misc info,
        youtube_id info, the file extension, and the extractions status value (True or False
        depending on whether data was sent to the `true_extract` method or the `false_extract`
        method by the Extractor).

        If the filename info is extracted successfully, the method uses the `true_extract` method
        to set the current extract data. If unsuccessful, the method uses the `false_extract` method
        to set the current extract data. Current extract data always contains the last extract
        within the current `Extractor()` instance. Extracts are tuples containing eight values:
        the original filename, the artist info, title info, features info, misc info, youtube_id
        info, the file extension, and the extractions status value which will be True or False
        depending on whether data was sent to the `true_extract` method or the `false_extract`
        method.

        Parameters:
            _file (LiteralString): The filename to extract information from.

        Returns:
            List[Any]: The extracted information in the form of a list containing eight values:
                the original filename, the artist info, title info, features info, misc info,
                youtube_id info, the file extension, and the extractions status value.
        """

        if not isinstance(_file, str):
            raise TypeError("file must be a str literal")

        _file = os.path.basename(_file)

        if not _file.endswith(self.exts):
            return self.false_extract(_file)

        self.simple_extract(_file)
        check = self.get_extraction("dict")
        if check["status"]:
            return

        # create a copy of the filename, one for editing, one for backup
        __file = _file

        # check for track features
        feature_regex_list = [
            (r"\([fF]t[\. | ](.+?)\)", 1),
            (r"([fF]t[\. | ]|[wW]\/)(.+?)(?=([\'\"\.]|[()]|[-]|[\[]))", 2),
            (r"\([fF]eat[\. | ](.+?)\)", 3),
            (r"([fF]eat[\. | ](.+?)(?=([\'\"\.]|[()]|[-]|[\[])))", 4),
        ]

        features = []

        for regex in feature_regex_list:

            features_match = re.search(regex[0], _file)
            group = 1 if regex[1]%2 else 2

            if features_match:
                _file = _file.replace(features_match[0].strip("-[("), "")
                features.append(features_match.group(group).strip("-[(").strip())

        # check for possible youtube id
        youtube_id_regex = re.search('[A-Za-z0-9_-]{11}', _file)
        youtube_id = None
        if youtube_id_regex:
            youtube_id = youtube_id_regex[0]
            _file = _file.replace(youtube_id, '').replace('()', '').replace('[]', '')

        # check for misc info
        misc_regex = re.findall(r"(\(.+?\))|(\[.+?\])|([pP]rod [bB ]y \w*)", _file)
        misc_list = []

        if misc_regex:
            for match in misc_regex:
                _match = match[0] if match[0] else None
                _match = match[1] if _match is None else _match
                _match = match[2] if _match is None else _match
                if _match:
                    _file = _file.replace(_match, "").replace("  ", " ").strip("()[] ")
                    misc_list.append(_match.strip("()[] "))

        # check for rest of values, first for a artist-title combo and then just for artist
        common_regex = re.search(
            r"^(.+?)[ ]?-[ ](.+?)\.(\w+)$|^(.+?)[ ]-[ ]?(.+?)\.(\w+)$|^(.+)\.(\w+)$",
            _file
        )

        # at this point if file cant be formatted, return with a false extract
        if not common_regex:
            return self.false_extract(__file)

        # if file is formattable, check for existing data and fill it in. Use defaults set in config
        # for cases where no info is available.
        features = ', '.join(features) if features else None
        misc = ", ".join(misc_list).strip("()") if misc_list else None

        common_groups = [
            (1, 3),
            (4, 6),
            (7, 8)
        ]

        for group in common_groups:

            if common_regex.group(group[0]):

                artist = (
                    common_regex.group(group[0]).strip()
                    if common_regex.group(group[0])
                    else None
                )

                filetype = (
                    common_regex.group(group[1]).strip().rstrip(".")
                    if common_regex.group(group[1])
                    else None
                )

        # if there is no title, double check the artist name to see if its possibly located there
        title = (
            common_regex.group(2).strip()
            if common_regex.group(2)
            else None
        )

        title = (
            common_regex.group(5).strip()
            if common_regex.group(5) and title is None
            else title
        )

        title_in_artist = re.search(r"([\uFF02\"\'\“\”].+?[\uFF02\"\'\“\”])|([：:•].+)", artist)

        if title_in_artist and title is None:

            if title_in_artist.group(1):
                artist = artist.replace(title_in_artist.group(1), "").strip()
                title = title_in_artist.group(1)

            else:
                artist = artist.replace(title_in_artist.group(2), "").strip()
                title = title_in_artist.group(2)

        if artist is None:
            artist = re.search(r"(.+?)\.(\w+)$", _file)
            artist = artist.group(1)

        # redundant file stripping
        if title is not None:
            title = title.strip("-：:•\uFF02\"'“ ")
        artist = artist.strip("-：:•\uFF02\"'“() ")

        # create formatted file name
        return self.true_extract([
                __file,
                artist,
                title,
                features,
                misc,
                youtube_id,
                filetype,
                True
        ])

    # extracts data from a list of files, allows selection of an output opt
    # which is set to "dict" by default. returns a list of extractions.
    # all extractions are tuples containing 8 values.
    def extract_complex_list(self, file_list: List[str], opt: str = "dict"):
        """
        Extracts data from a list of files using the `complex_extract` method from the `extract` module. 
        Allows selection of an output option, which is set to "dict" by default. Returns a list of 
        extractions, where each extraction is a tuple containing 8 values.

        get_extraction method not necessary to call after use of this method, a list of extractions is
        returned. get_extraction in this case would return the last extraction preformed.

        Parameters:
            file_list (List[str]): A list of file paths from which data is to be extracted.
            opt (str, optional): 
                An output option that determines the format of the extracted data. Defaults to "dict". 
                The supported options are "dict", "list", and "tuple".

        Returns:
            List: A list of extractions, where each extraction is a tuple containing 8 values.

        Raises:
            TypeError: 
                If the `file_list` parameter is not a list of strings or if the `opt` parameter
                is not a string corresponding to the supported options.
        """
        _extractions = []
        if isinstance(file_list, list) and isinstance(opt, str):
            for _file in file_list:
                self.complex_extract(_file)
                _extractions.append(self.get_extraction(opt.strip().lower()))
            return _extractions

        raise TypeError(
            "File_list must be a list of strings. Opt should be a string corresponding to output options"
        )



import os
import configparser
import pkg_resources


class ConfigBase:
    """
    Package config values are set, these should not be changed by user
    """
    def __init__(self):
        self.config = configparser.ConfigParser()
        self.userconfig = configparser.ConfigParser()
        self.path = pkg_resources.resource_filename(__name__, "config.ini")
        self.config.read(self.path)

    @property
    def user_paths(self):
        try:
            paths = self.config['PROGRAM']['userpaths'].replace(' ', '').split(',')
            return [os.path.expanduser(path.strip()) for path in paths]
        except:
            raise KeyError('PACKAGE CONFIG HAS ERRORS')

    @property
    def app_name(self):
        try:
            return self.config['APP']['name']
        except:
            raise KeyError('PACKAGE CONFIG HAS ERRORS')
    
    @property
    def output_opts(self):
        try:
            return self.config['PROGRAM']['output_opts'].replace(' ', '').split(',')
        except:
            raise KeyError('PACKAGE CONFIG HAS ERRORS')

    @property
    def constructors(self):
        try:
            return self.config['PROGRAM']['constructors'].replace(' ', '').split(',')
        except:
            raise KeyError('PACKAGE CONFIG HAS ERRORS')

class ConfigUser(ConfigBase):
    """
    Configurable user settings
    """
    def __init__(self, config_path: str = None):
        super().__init__()
        if config_path is not None:
            self.userconfig.read(config_path)
            self.path = config_path
        else:
            for path in self.user_paths:
                if os.path.exists(path):
                    self.userconfig.read(path)
                    self.path = path
                    break
            if self.path is None:
                self.path = pkg_resources.resource_filename(__name__, "config.ini")

    @property
    def config_path(self):
        if self.path is None:
            raise KeyError("CONFIG PATH ERROR")
        return self.path

    @property
    def db_path(self):
        try:
            return self.userconfig['DATABASE']['path']
        except KeyError:
            try:
                return self.config['DATABASE']['path']
            except:
                raise TypeError("PROBLEM WITH CONFIG")

    @property
    def dry(self):
        try:
            dry = self.userconfig['PROGRAM']['dry']
            return dry.lower() == 'true'
        except KeyError:
            try:
                dry = self.config['PROGRAM']['dry']
                return dry.lower() == 'true'
            except:
                raise TypeError("PROBLEM WITH CONFIG")

    @property
    def exts(self):
        try:
            return self.userconfig['PROGRAM']['exts'].replace(' ', '').split(',')
        except KeyError:
            try:
                return self.config['PROGRAM']['exts'].replace(' ', '').split(',')
            except:
                raise TypeError("PROBLEM WITH CONFIG")



"""
Basically used for all print handling
"""
import os
from rich.console import Console
from rich.theme import Theme
from rich.progress import track
from audiodotturn import VERSION
from audiodotturn import AudioDotTurn
from audiodotturn.parser import Parser


def init():
    """
    Initialize AudioDotTurn.

    Returns:
        tuple:
            args (parser.parse_args()),
            adt (AudioDotTurn)
    """
    parser = Parser()
    args = parser.parse_args()
    config_path = args.cfgpath or None
    db_path = args.dbpath or None
    adt = AudioDotTurn(config_path=config_path, db_path=db_path)

    return args, adt

def rich_inits():
    #rich initialisations
    theme = Theme(
        {
            "success": "bold green",
            "failure": "magenta",
            "error": "bold red",
            "info": "bold yellow"
        }
    )
    console = Console(theme=theme)

    return console

def extract_commands(args, adt: AudioDotTurn):
    """
    Extracts information from a given file or multiple files.

    Args:
        args (Namespace): Command line arguments.
        adt (AudioDotTurn): AudioDotTurn instance.
    """
    console = rich_inits()
    opt = args.out
    success = 0
    failure = 0

    if args.file:
        extraction = adt.extract_file(args.file, opt)[0]
        console.print(extraction, '\n', style="info")

    elif args.multi:
        extractions = adt.extract_files(args.multi, opt)
        adt.produce_extract_report(extractions, console)
        for extracted in extractions:
            if extracted["status"]:
                success += 1
            else:
                failure += 1
        console.print(f"\nsuccess: {success}\nfailure: {failure}")

    elif args.dir:
        try:
            files = [file for file in os.listdir(args.dir) if os.path.isfile(os.path.join(args.dir, file)) and file.endswith(tuple(adt.config.exts))]
        except NotADirectoryError as error:
            console.print(error)
            return

        extractions = adt.extract_files(files, opt)
        adt.produce_extract_report(extractions, console)
        for extracted in extractions:
            if extracted["status"]:
                success += 1
            else:
                failure += 1
        console.print(f"\nsuccess: {success}\nfailure: {failure}")


def construct_commands(args, adt: AudioDotTurn):
    """
    Constructs an output file from a given file or multiple files.

    Args:
        args (Namespace): Command line arguments.
        adt (AudioDotTurn): AudioDotTurn instance.
    """
    console = rich_inits()
    constructor_type = args.constructor
    success = 0
    failure = 0

    if args.file:
        extraction = adt.extract_file(args.file)
    elif args.multi:
        extraction = adt.extract_files(args.multi)
        adt.produce_extract_report(extraction, console)
    elif args.dir:
        extraction = adt.extract_file([file for file in os.listdir(args.dir) if os.path.isfile(os.path.join(args.dir, file)) and file.endswith(tuple(adt.config.exts))])
        adt.produce_extract_report(extraction, console)

    if not any((args.file, args.multi)):
        return

    results = adt.construct(constructor_type, extraction, args.auto)

    success, failure = adt.produce_construct_report(results, console, args)

    if success == 0 and failure == 0:
        console.print("\n[bold red]Run aborted.")
    else:
        console.print(f"\nsuccess: {success}\nfailure: {failure}")


def database_commands(args, adt: AudioDotTurn):
    """
    Manipulates the database.

    Args:
        args (Namespace): Command line arguments.
        adt (AudioDotTurn): AudioDotTurn instance.
    """
    console = rich_inits()

    if args.updatefile:

        console.print(f"Current database in use or to be created is {os.path.abspath(adt.config.db_path)}\n")
        confirm = input("continue? [y/N]")
        if confirm.lower() not in ['yes', 'y', 'yy']:
            console.print("Exiting\n", style="error")
            return

        extraction = adt.extract_file(args.updatefile)

        console.print('Extracted Info:', style="cyan")
        for key, value in extraction[0].items():
            console.print(key, ':', value, style="success")
        console.print('\n')

        confirm = input("update database [y/N]: ")
        if confirm.lower() not in ['yes', 'y', 'yy']:
            console.print("Exiting\n", style="error")
            return

        new_artists, new_songs, updated, failure = adt.update_database(extraction)
        if new_artists:
            console.print("New artist successfully added\n")
        if new_songs:
            console.print("New song successfully added\n")
        if updated:
            console.print("Entry updated successfully\n")
        if failure:
            console.print("Update failed\n")
    
    elif args.updatemulti:

        console.print(f"Current database in use or to be created is {os.path.abspath(adt.config.db_path)}\n", style="cyan")
        confirm = input("continue? [y/N]")
        if confirm.lower() not in ['yes', 'y', 'yy']:
            console.print("Exiting\n", style="error")
            return

        confirm_print = input("\nExamine extraction results before updating? [y/N]")
        if confirm_print.lower() not in ['yes', 'y', 'yy']:
            console.print("Exiting\n", style="error")
            return

        extractions = adt.extract_files(args.updatemulti)

        adt.produce_report(extractions, console)
                
        confirm = input("update database [y/N]: ")
        if confirm.lower() not in ['yes', 'y', 'yy']:
            console.print("Exiting\n", style="error")
            return

        new_artists, new_songs, updated, failure = adt.update_database(extractions)
        console.print(
            f"New artists: {new_artists}",
            f"New songs: {new_songs}\n",
            f"Updated: {updated}",
            f"Failure: {failure}\n"
        )

    elif args.artists:
        artists = adt.get_all_artists()
        if artists:
            for artist in track(artists, "Fetching artists..."):
                console.print(f'id: {artist["artist_id"]}, name: {artist["name"]}')
        else:
            console.print('None found.\n', style="info")

    elif args.songs:
        songs = adt.get_all_artists_and_songs()
        if songs:
            for artist, songs in track(songs.items(), "Fetching tracks..."):
                console.print(artist, style="success")
                for _id, song in enumerate(songs, start=1):
                    console.print(str(song).strip('{}'))
        else:
            console.print('None found.\n', style="info")

    elif args.artistid:
        songs = adt.get_songs_by_artist(args.artistid)
        if songs:
            for song in track(songs, "Fetching tracks..."):
                console.print(song, style="success")
        else:
            console.print('None found.\n', style="info")

    elif args.songid:
        song = adt.get_song_by_id(args.songid)
        if song:
            console.print(song, style="success")
        else:
            console.print("None found.\n", style="info")


def main():
    """
    Main function.
    """
    try:
        args, adt = init()
        console = rich_inits()

        if args.version:
            console.print(adt.config.app_name, VERSION, style="cyan")

        if args.settings:
            settings = {
                "audiodoturn": VERSION,
                "config path": adt.config.config_path,
                "db path": adt.config.db_path,
                "constructors": adt.config.constructors,
                "exts": adt.config.exts,
                "output options": adt.config.output_opts,
                "dry run": adt.config.dry
            }

            for key, value in settings.items():
                console.print(key, ':', value)
        
        if args.command == "extract":
            extract_commands(args, adt)

        elif args.command == "construct":
            construct_commands(args, adt)

        elif args.command == "database":
            database_commands(args, adt)

    except Exception as error:
        print(error)

if __name__ == "__main__":
    main()



import argparse

class Parser:
    def __init__(self):
        """
        Initializes the parser with arguments and sub-commands.

        Returns:
            None
        """
        self.parser = argparse.ArgumentParser(description='Format, organize and retrieve data from audio files.')
        
        # Add top-level arguments
        self.parser.add_argument('-v', '--version', action='store_true', help='Show current version of audiodotturn')
        self.parser.add_argument('-p', '--cfgpath', help='Path to a specific configuration file to use for the session.')
        self.parser.add_argument('-d', '--dbpath', help='Path to .db file for library database')
        self.parser.add_argument('-s', '--settings', action='store_true', help='Show current settings')
        
        # Add subparsers
        self.subparsers = self.parser.add_subparsers(dest='command')
        
        # Create parser for the "extract" command
        self.extract_parser = self.subparsers.add_parser('extract', help='Extraction commands')
        self.extract_parser.add_argument('-o', '--out', default="dict",type=str, help='Output format for extraction, default is dict.')
        self.extract_parser.add_argument('-f', '--file', type=str, help='Extract info from single file.')
        self.extract_parser.add_argument('-m', '--multi', nargs="+", type=str, help='Extract info from multiple files.')
        self.extract_parser.add_argument('-l', '--dir', type=str, help='Extract info from files in a directory.')
        
        # Create parser for the "construct" command
        self.construct_parser = self.subparsers.add_parser('construct', help='Construction commands')
        self.construct_parser.add_argument('-a', '--auto', action="store_true", help='Set auto-choice')
        self.construct_parser.add_argument('-c', '--constructor', default="simple", help='Constructor to use')
        self.construct_parser.add_argument('-f', '--file', help='Construct from a single file')
        self.construct_parser.add_argument('-m', '--multi', nargs="+", help='Construct from multiple files')

        # Create parser for the "database" commands
        self.database_parser = self.subparsers.add_parser('database', help='Database commands')
        self.database_parser.add_argument('-f', '--updatefile', help="Update database via file.")
        self.database_parser.add_argument('-m', '--updatemulti', nargs="+", help="Update database via multiple files.")
        self.database_parser.add_argument('-A', '--artists', action="store_true", help='View all artists within the database')
        self.database_parser.add_argument('-S', '--songs', action="store_true", help='View all songs by each artist within the database')
        self.database_parser.add_argument('-Ai', '--artistid', type=int, help='View songs by artist id')
        self.database_parser.add_argument('-Si', '--songid', type=int, help='View song by song id')

    def get_parsers(self):
        """
        Return a list of parsers that can be used to parse command-line arguments.
            parser, create_parser, view_parser, set_parser
        """
        parsers = [self.parser, self.extract_parser, self.construct_parser, self.database_parser]
        return parsers
    
    def parse_args(self, args=None):
        """
        Parse command-line arguments using the parser.

        Args:
            args (list): The arguments to parse. If None, command-line arguments are used.

        Returns:
            Namespace: argparse.Namespace object of the parsed arguments.
        """
        return self.parser.parse_args(args)


import sqlite3
from typing import List, Dict

class DatabaseInit:
    def __init__(self, path: str) -> None:
        """
        Constructs a new Database object.

        Parameters:
            path : str
                A string representing the file path of the database. Database must be a `.db` file.
        """
        self.path = path if path.endswith('.db') else None
        if self.path is None:
            raise TypeError("Database must be a .db file")


class DatabaseCreate(DatabaseInit):
    def create_database(self) -> None:
        """
        Creates a new database file at the specified path if it does not exist.
        """
        try:
            conn = sqlite3.connect(self.path)
            conn.close()
        except:
            raise sqlite3.OperationalError(f'Could not create database in {self.path}')

    def create_tables(self) -> None:
        """
        Creates the necessary tables for the database.
        """
        conn = sqlite3.connect(self.path)
        cursor = conn.cursor()

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS artists (
                artist_id INTEGER PRIMARY KEY,
                name TEXT UNIQUE
            )
        """)

        cursor.execute("""
            CREATE TABLE IF NOT EXISTS songs (
                song_id INTEGER PRIMARY KEY,
                title TEXT,
                features TEXT,
                misc TEXT,
                youtube_id TEXT,
                file_extension TEXT,
                artist_id INTEGER,
                FOREIGN KEY(artist_id) REFERENCES artist(artist_id)
            )
        """)

        conn.commit()
        conn.close()


class DatabaseRead(DatabaseInit):
    def get_all_artists(self) -> List[Dict]:
        """
        Retrieves a list of all artists in the database.

        Returns:
            A list of dictionaries representing each artist in the database,
            with keys 'artist_id' and 'name'.
        """
        conn = sqlite3.connect(self.path)
        cursor = conn.cursor()
        cursor.execute('SELECT artist_id, name FROM artists')
        result = cursor.fetchall()
        conn.close()
        if result:
            return [{'artist_id': row[0], 'name': row[1]} for row in result]
        return None

    def get_artist_by_id(self, artist_id: int) -> Dict:
        """
        Retrieves the artist with the given ID from the database.

        Parameters:
            artist_id : int
                The ID of the artist to retrieve.

        Returns:
            A dictionary representing the artist in the database,
            with keys 'artist_id' and 'name'.
        """
        conn = sqlite3.connect(self.path)
        cursor = conn.cursor()
        cursor.execute('SELECT artist_id, name FROM artists WHERE artist_id = ?', (artist_id,))
        result = cursor.fetchall()
        conn.close()
        if result:
            return [{'artist_id': row[0], 'name': row[1]} for row in result]
        return None

    def get_songs_by_artist_by_id(self, artist_id: int) -> List[Dict]:
        """
        Retrieves a list of all songs by the artist with the given ID from the database.

        Parameters:
            artist_id : int
                The ID of the artist whose songs to retrieve.

        Returns:
            A list of dictionaries representing each song by the artist in the database,
            with keys 'song_id', 'title', 'features', 'misc', 'youtube_id', and 'file_extension'.
        """
        conn = sqlite3.connect(self.path)
        cursor = conn.cursor()
        cursor.execute('SELECT song_id, title, features, misc, youtube_id, file_extension '
                    'FROM songs WHERE artist_id = ?', (artist_id,))
        result = cursor.fetchall()
        conn.close()
        if result:
            return [{'song_id': row[0], 'title': row[1], 'features': row[2],
                    'misc': row[3], 'youtube_id': row[4], 'file_extension': row[5]} for row in result if row]
        return None

    def get_song_by_id(self, song_id: int) -> Dict:
        """
        Retrieves the song with the given ID from the database.

        Parameters:
            song_id : int
                The ID of the song to retrieve.

        Returns:
            A dictionary representing the song in the database,
            with keys 'song_id', 'title', 'features', 'misc', 'youtube_id', and 'file_extension'.
        """
        conn = sqlite3.connect(self.path)
        cursor = conn.cursor()
        cursor.execute('SELECT song_id, title, features, misc, youtube_id, file_extension '
                    'FROM songs WHERE song_id = ?', (song_id,))
        result = cursor.fetchall()[0]
        conn.close()
        if result:
            return {'song_id': result[0], 'title': result[1], 'features': result[2],
                    'misc': result[3], 'youtube_id': result[4], 'file_extension': result[5]}
        return None


class DatabaseUpdate(DatabaseInit):
    def update_database(self, data: List[Dict]) -> None:
        """
        Updates the database with the provided data from an extractor.
        If the artist exists, the data is added as a new song for that artist.
        If the artist does not exist, a new artist entry is created and the data is added as the first song.
        If the song already exists, any missing information is updated without changing existing information.

        Parameters:
            data : list of dicts
                A list of dicts representing the extracted music metadata:
                the original filename, the artist info, title info, features info, misc info, youtube_id info,
                the file extension, and the extraction status value (True or False).
        Returns:
            tuple : (new_artists, new_songs, updated, failure)
                Stats of last update run
        """
        conn = sqlite3.connect(self.path)
        cursor = conn.cursor()
        new_artists = 0
        new_songs = 0
        updated = 0
        failure = 0


        for row in data:
            if not row["status"]:
                failure += 1
                continue

            artist_name = row["artist"].lower()
            song_title = row["title"]
            features = row["features"]
            misc = row["misc"]
            youtube_id = row["youtube_id"]
            file_extension = row["filetype"]

            # check if artist exists in database
            try:
                cursor.execute('SELECT artist_id FROM artists WHERE name = ?', (artist_name,))
                result = cursor.fetchone()

            except sqlite3.OperationalError:
                result = False

            if result:
                artist_id = result[0]
                cursor.execute('SELECT song_id, features, misc, youtube_id FROM songs WHERE title = ? AND artist_id = ?',
                            (song_title, artist_id))
                result = cursor.fetchone()
                if result:
                    check = False
                    song_id = result[0]
                    # update only the missing information for the existing song
                    if result[1] is None:
                        check = True
                        cursor.execute('UPDATE songs SET features = ? WHERE song_id = ?', (features, song_id))
                    if result[2] is None:
                        check = True
                        cursor.execute('UPDATE songs SET misc = ? WHERE song_id = ?', (misc, song_id))
                    if result[3] is None:
                        check = True
                        cursor.execute('UPDATE songs SET youtube_id = ? WHERE song_id = ?', (youtube_id, song_id))
                    if check:
                        updated += 1
                else:
                    # add a new song for the artist
                    cursor.execute('INSERT INTO songs (artist_id, title, features, misc, youtube_id, file_extension) '
                                'VALUES (?, ?, ?, ?, ?, ?)', (artist_id, song_title, features, misc, youtube_id, file_extension))
                    new_songs += 1
            else:
                # add a new artist and song
                cursor.execute('INSERT INTO artists (name) VALUES (?)', (artist_name,))
                artist_id = cursor.lastrowid
                cursor.execute('INSERT INTO songs (artist_id, title, features, misc, youtube_id, file_extension) '
                            'VALUES (?, ?, ?, ?, ?, ?)', (artist_id, song_title, features, misc, youtube_id, file_extension))
                new_songs += 1
                new_artists += 1

        conn.commit()
        conn.close()
        return new_artists, new_songs, updated, failure


class Database(DatabaseCreate, DatabaseRead, DatabaseUpdate):
    """
    Subclass of all Database Classes, usually what will be instantiated.
    """
    def __init_subclass__(cls) -> None:
        return super().__init_subclass__()
